/**
 * @file		stringfmt.cc
 * @author	daniel
 * @brief
 */

#include <string>
#include <iostream>

#include <type_traits>
#include <tuple>
#include <utility>
#include <cstdint>

// just for printing type information
#include <typeinfo>
#include <cxxabi.h>
#include <memory>

//template<typename ...>
//class printType;


/// for simplicity, all serialization goes into a string
using buffer_type = std::string;


class MyIpV6Addr;


/**
 * Placeholder for string-like objects (char*, std::string etc.)
 */
struct str
{
   /// mark this as a placeholder
   using placeholder = std::true_type;

   /// specified for all types that are acceptable as arguments
   template<typename T> struct accepts : public std::false_type {};

   /**
    * Actual string formatting implementation: appends to the buffer.
    * @param[out] buffer      the buffer to append to
    * @param[in]  arg         the value to append
    */
   template<typename Arg>
   void format(buffer_type& buffer, Arg&& arg) const
   {
      buffer.append(arg);
   }
};
// the types we accept as arguments
template<> struct str::accepts<char*> : public std::true_type {};
template<> struct str::accepts<const char*> : public std::true_type {};
template<> struct str::accepts<std::string> : public std::true_type {};


/**
 * Placeholder for numbers (int, long, ...). Currently, only decimal output is implemented.
 */
struct num
{
   // TODO: width, precision, hex etc.

   /// mark this as a placeholder
   using placeholder = std::true_type;

   /// specified for all types that are acceptable as arguments
   template<typename T> struct accepts : public std::false_type {};

   /**
    * Actual string formatting implementation: appends the decimal representation of the argument
    * to the buffer.
    * @param[out] buffer      the buffer to append to
    * @param[in]  arg         the numeric value to append
    */
   template<typename Arg>
   void format(buffer_type& buffer, Arg arg) const
   {
      // TODO: could be more efficient...
      buffer.append(std::to_string(arg));
   }
};
// the types we accept as arguments
template<> struct num::accepts<int8_t> : public std::true_type {};
template<> struct num::accepts<uint8_t> : public std::true_type {};
template<> struct num::accepts<int16_t> : public std::true_type {};
template<> struct num::accepts<uint16_t> : public std::true_type {};
template<> struct num::accepts<int32_t> : public std::true_type {};
template<> struct num::accepts<uint32_t> : public std::true_type {};
template<> struct num::accepts<int64_t> : public std::true_type {};
template<> struct num::accepts<uint64_t> : public std::true_type {};
#ifndef _LP64
template<> struct num::accepts<long> : public std::true_type {};
template<> struct num::accepts<unsigned long> : public std::true_type {};
#endif


/// Example specialization for 'high performance' serialization of small integers
template<>
void num::format<uint8_t>(buffer_type& buffer, uint8_t arg) const
{
   // range: 0 - 256
   // -> format each digit individually
   char buf[4]{0, 0, 0, 0};
   buf[2] = '0' + arg % 10;
   buf[1] = '0' + (arg % 100) / 10;
   buf[0] = '0' + (arg / 100);

   // skip leading zeros
   size_t offset = 0;
   while (buf[offset] == '0' && offset < 2)
      ++offset;
   buffer += buf + offset;
}
template<>
void num::format<int8_t>(buffer_type& buffer, int8_t arg) const
{
   std::cout << " -- using the fast algorithm!\n";
   // check for negative numbers, then process as if it where signed
   if (arg < 0)
   {
      buffer += '-';
      arg = -arg;
   }
   format(buffer, static_cast<uint8_t>(arg));
}


/**
 * Placeholder for pointers.
 */
struct ptr
{
   /// mark this as a placeholder
   using placeholder = std::true_type;

   /// specified for all types that are acceptable as arguments
   template<typename T> struct accepts : public std::is_pointer<T> {};

   /**
    * Actual string formatting implementation: appends to the buffer.
    * @param[out] buffer      the buffer to append to
    * @param[in]  arg         the value to append
    */
   template<typename Arg>
   void format(buffer_type& buffer, Arg arg) const
   {
      // yeah, bad...
      char tmp[32];
      snprintf(tmp, sizeof(tmp), "%p", static_cast<const void*>(arg));
      buffer.append(tmp);
   }
};
// it seems that nullptr is not a pointer...
template<> struct ptr::accepts<nullptr_t> : public std::true_type {};


// TODO: add more placeholders, e.g. for datetime objects, float/double, pointers etc.


/**
 * Type traits that check if a type is a placeholder.
 */
// default: no
template<typename T, typename Enable=void>
struct is_placeholder: public std::false_type
{ };
// otherwise: check for the contained 'placeholder' type
template<typename T>
struct is_placeholder<T, typename std::enable_if<std::remove_reference_t<T>::placeholder::value>::type>
                     : public std::true_type
{ };

// make sure we didn't screw this up
static_assert(is_placeholder<str>::value, "oops");
static_assert(is_placeholder<const str>::value, "oops");
static_assert(is_placeholder<str&>::value, "oops");
static_assert(is_placeholder<const str&>::value, "oops");
static_assert(not is_placeholder<int>::value, "oops");
static_assert(not is_placeholder<std::string>::value, "oops");


/*
 * We build some helpers that will allow us to extract placeholders from a type list. This
 * is necessary in order to verify if the parameters passed to the format string actually
 * match the placeholders.
 *
 * The implementation is strait forward:
 * Process the type list element by element and build a new tuple. If the current type isn't
 * a placeholder, skip it. Continue recursively.
 */
template<typename...>
struct filter_placeholders;

// stop when the type list is empty
template<> struct filter_placeholders<>
{ using type = std::tuple<>; };


// helper to "merge" a type into a tuple to flatten the type list again
template <typename, typename> struct merge_tuple;

template <typename  T, typename ...Args>
struct merge_tuple<T, std::tuple<Args...>>
{
    using type = std::tuple<T, Args...>;
};

template<typename Head, typename ... Tail>
struct filter_placeholders<Head, Tail...>
{
   using type = typename std::conditional<
         is_placeholder<Head>::value,
         // placeholder: keep, check the rest
         typename merge_tuple<Head, typename filter_placeholders<Tail...>::type>::type,
         // no placeholder: drop it, check the rest
         typename filter_placeholders<Tail...>::type
         >::type;
};


/*
 * Templates to check a list of arguments against the list of placeholders.
 * Both are given as tuples so that both tuples can be checked element for element.
 */

template<typename Tuple1, typename Tuple2, int Index>
struct argument_check;

// end of recursion: both tuples are empty, all checks succeeded
template<int Index>
struct argument_check<std::tuple<>, std::tuple<>, Index> : public std::true_type {};

// process recursively by checking index I in this type and inheriting from I+1's check
template<typename T1, typename ... R1, typename T2, typename ... R2, int Index>
struct      argument_check<std::tuple<T1, R1...>, std::tuple<T2, R2...>, Index>
   : public argument_check<std::tuple<    R1...>, std::tuple<    R2...>, Index + 1>
{
   using Placeholder = T1;
   using ArgType = T2;
   // verify that the argument's type is accepted by the placeholder
   static_assert(Placeholder::template accepts<ArgType>::value == true,
                 "Invalid argument type(s)");
};


/*
 * Implementation of the string formatting: The argument type checks have already passed,
 * otherwise the error messages would be pretty nasty...
 */

template<typename Args, size_t index, bool placeholder, typename Spec, typename ... Specs>
struct _format_impl;


// dummy for an empty spec list (end of iteration)
template<typename Args, size_t index>
void _format_next(buffer_type& result, const Args& args)
{
}

// format spec by spec (recursively), using the arguments as needed
template<typename Args, size_t index, typename Spec, typename ... Specs>
void _format_next(buffer_type& result, const Args& args, const Spec& s, Specs&&... specs)
{
   // pick the right formatter (placeholder or fixed)
   using Formatter = _format_impl<Args, index, is_placeholder<Spec>::value, Spec, Specs...>;
   Formatter::format(result, args, s, specs...);
}


/**
 * Formatter for placeholders: Uses the argument at the current index to call the
 * placeholder's format function. The current index and spec are 'consumed' and
 * _format_next is called again.
 */
template<typename Args, size_t index, typename Spec, typename ... Specs>
struct _format_impl<Args, index, true, Spec, Specs...>
{
   static void format(buffer_type& result, const Args& args, const Spec& s, Specs&&... specs)
   {
      static_assert(is_placeholder<Spec>::value, "oops");
      static_assert(index < std::tuple_size<Args>::value, "oops");
      // format using the arg
      s.format(result, std::get<index>(args));
      // process the rest (the index is incremented and the current spec is dropped)
      _format_next<Args, index + 1, Specs...>(result, args, specs...);
   }
};

/**
 * Formatter for fixed strings (non-placeholders): The arguments remain untouched.
 * The current spec is formatted and _format_next is called again.
 */
template<typename Args, size_t index, typename Spec, typename ... Specs>
struct _format_impl<Args, index, false, Spec, Specs...>
{
   static void format(buffer_type& result, const Args& args, const Spec& s, Specs&&... specs)
   {
      static_assert(not is_placeholder<Spec>::value, "oops");
      // note: the index may already be past the end if all args have been used
      static_assert(index <= std::tuple_size<Args>::value, "oops");
      // format with any arg
      s.format(result);
      // process the rest
      _format_next<Args, index, Specs...>(result, args, specs...);
   }
};


template<typename Specs, typename Args, std::size_t... I>
void _format(buffer_type& result, const Args& args, const Specs& specs,
             const std::index_sequence<I...>)
{
   // unpack the specs from the tuple into an argument pack
   _format_next<Args, 0>(result, args, std::get<I>(specs)...);
}


/**
 * Format-class: type parameters can be wrappers for fixed values or placeholders. They fixed
 * values are copied into the object. The object can be called with all required arguments
 * to generate a completely formatted string.
 */
template<typename ... Specs>
struct fmt
{
   // list of all format specifications
   typedef std::tuple<Specs...> SpecTypes;
   // list of all variable parameters that are required to be passed to operator()
   typedef typename filter_placeholders<Specs...>::type ParamTypes;

   SpecTypes _specs;

   fmt(Specs... specs)
   : _specs(std::move(specs)...)
   {
   }

   /// helper struct to include size information in compiler error messages (because they are part of the type)
   template<size_t X, size_t Y>
   struct equal_size: public std::integral_constant<bool, X == Y>
   {
      equal_size()
      {
         static_assert(X == Y, "Invalid number of arguments");
      }
   };


   /**
    * Call operator: Takes the arguments required for all placeholders, formats everything
    * and returns the formatted string.
    * This function tries to give meaningful and readable error messages in case that the
    * arguments passed to this function don't match the specified placeholders.
    */
   template<typename ... FmtArgs>
   buffer_type operator()(FmtArgs&&... args) const
   {
      // remove all qualifiers from the argument types, decay character arrays into pointers
      using ArgumentTypes = std::tuple<typename std::remove_cv<typename std::decay<FmtArgs>::type>::type...>;
      // make sure that the number of arguments matches the number of placeholders
      equal_size<std::tuple_size<ArgumentTypes>::value, std::tuple_size<ParamTypes>::value> _size_check;
      (void)_size_check;

      // check the argument types against the types that the placeholders support
      static_assert(argument_check<ParamTypes, ArgumentTypes, 0>::value == true, "!?!");

      buffer_type result;
      auto argsTuple = std::make_tuple(std::forward<FmtArgs>(args)...);
      _format(result, argsTuple, _specs,
              std::make_index_sequence<std::tuple_size<SpecTypes>::value>());
      return result;
   }
};


/**
 * Wrapper type for fixed parts of a format string (currently only strings).
 */
template<typename T>
struct wrapper_t
{
   const T value; // usually: const char*
   wrapper_t(T v): value(v) {}

   wrapper_t(const wrapper_t&) = default;
   wrapper_t(wrapper_t&&) = default;
   wrapper_t& operator=(const wrapper_t&) = default;
   wrapper_t& operator=(wrapper_t&&) = default;
   ~wrapper_t() = default;

   void format(buffer_type& buffer) const
   {
      buffer.append(value);
   }
};

// make sure we don't accidentally mix up fixed strings with string-placeholders
static_assert(not is_placeholder<wrapper_t<std::string>>::value, "oops");

/*
 * Wrapper traits: for fixed strings, use the wrapper_t<> template, but keep placeholders as-is
 */

template<typename T, typename Enable=void>
struct wrap_traits { using type = T; };

template<typename T>
struct wrap_traits<T, typename std::enable_if<   std::is_integral<T>::value
                                              or std::is_pointer<T>::value>::type>
{ using type = wrapper_t<T>; };

template<typename T>
auto wrap_literal(T x)
{
   using wrapped_type = typename wrap_traits<T>::type;
   return wrapped_type(std::move(x));
}


/**
 * Constructs the actual format string instance from the given arguments.
 * All passed arguments must be either wrapper literals/strings or placeholders.
 */
template<typename ... Args>
fmt<Args...> _build_fmt(Args... args)
{
   return fmt<Args...>(args...);
}

/**
 * Build a format string from the given list of fixed strings an placeholder constructs.
 */
template<typename ... Args>
auto build_format_string(Args... args)
{
   return _build_fmt(wrap_literal(std::move(args))...);
}



static std::string demangle(const char* name)
{
   int status = -1;
   std::unique_ptr<char, void (*)(void*)> res {
      abi::__cxa_demangle(name, NULL, NULL, &status),
      std::free
   };
   return (status == 0) ? res.get() : name;
}

int main()
{
   const char* x = "x";

   auto f = build_format_string("Hello ", str(), " #", num(), "!\n");
   std::cout << f((const char*)"world", 5);
   std::cout << f("world", 7L);
   std::cout << f(x, (unsigned int)88);
   std::cout << f(x, (int8_t)-127);
   std::cout << f("world", -1);
   std::cout << f("world", (uint8_t)9);
   // these won't compile
//   std::cout << f();
//   std::cout << f(1, 2);
//   std::cout << f(5);

   auto f1 = build_format_string("Pointer value: ", ptr());
   std::cout << f1(&f1) << "\n";
   std::cout << f1(x) << "\n";
   std::cout << f1(nullptr) << "\n";

   auto f2 = build_format_string();
   std::cout << "'" << f2() << "'\n";

   auto f3 = build_format_string("Constant string");
   std::cout << "'" << f3() << "'\n";

   std::cout << "\n"
         << "sizeof(f) : " << sizeof(f ) << " (" << demangle(typeid(f ).name()) << ")\n"
         << "sizeof(f2): " << sizeof(f2) << " (" << demangle(typeid(f2).name()) << ")\n"
         << "sizeof(f3): " << sizeof(f3) << " (" << demangle(typeid(f3).name()) << ")\n";

   return 0;
}

